// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.IO;
using System.Net;
using System.Xml;
using System.Text.RegularExpressions;
using BoPi.Common;

namespace BoPi.Compiler
{
  /// <summary>
  /// Compiles BoPi programs producing the XML bytecode
  /// </summary>
  class Compiler
  {

    [STAThread]
    static void Main(string[] args)
    {			
      String m_input = "";
      Process ast = null;
      BoPiParser prg = new BoPiParser();		
      String host = "", filename;
			
      if (args.Length < 1) {
	  		Console.WriteLine("Usage: BoPiCompiler <filename>");
	  		return;
			}		

      if (args[0].IndexOf("@") != -1) {
	  		host = args[0].Substring(args[0].IndexOf("@") + 1);
	  		filename = args[0].Substring(0, args[0].IndexOf("@"));
			} else
				filename = args[0];

			CompilerOutput.FileName = filename;

      try {
	  		using (StreamReader file = new StreamReader(filename)) {
	      	while (file.Peek() != -1)
		  			m_input += file.ReadLine() + "\n";				
	      	file.Close();
	    	}
			} catch (Exception) {
	  		Console.WriteLine("Error reading file {0}", filename);
	  		return;
			}

      try {
    		string any = "any = void + int + string + (~[any],any) + <empty>^O + <any>^I";
    		string empty = "empty = ~[empty]";
 	  		m_input = "schemadef " + any + " and " + empty + " in " + m_input;			
	  		ast=(Process) prg.parse(m_input);
				ast.ResolveNames(new CompilerTypeSymbolTable(), new CompilerValueSymbolTable());
			} catch (Exception e) {
	  		Console.WriteLine(e.Message);
	  		return;
			}

      ast.Check();
      if (CompilerOutput.HasErrors) {
	  		CompilerOutput.PrintResult();
	  		return;
			}

      XmlTextWriter dec;
      StringWriter swrdec = null;
      StringWriter swr = new StringWriter();
      XmlTextWriter code = new XmlTextWriter(swr);
      code.Formatting = XmlUtil.IndentMode;
      if (host != "") {                
	  		if (!host.StartsWith("http://")) {
	      	CompilerOutput.Message("Must specify a valid location to upload code");
	      	return;
	    	}

	  		swrdec = new StringWriter();
	  		dec = new XmlTextWriter(swrdec);
	  		dec.Formatting = XmlUtil.IndentMode;
			} else {
	  		dec = new XmlTextWriter(args[0] + ".xml", System.Text.Encoding.UTF8);            
	  		dec.Formatting = XmlUtil.IndentMode;
	  		dec.WriteStartDocument();
			}

      dec.WriteStartElement("BoPi");
      dec.WriteAttributeString("xmlns", null, null, "http://cs.unibo.it/BoPi/opcode");
      dec.WriteAttributeString("xmlns", "value", null, "http://cs.unibo.it/BoPi/value");
      dec.WriteStartElement("schemadecl");
			AllocateContext allocateContext = new AllocateContext();
			ast.AllocateNames(allocateContext);
      ast.Compile(dec, code);
      dec.WriteEndElement();
      dec.WriteStartElement("process");			
			dec.WriteAttributeString("env-size", allocateContext.Size.ToString());
      XmlUtil.CopyXml(dec, swr.ToString());
      dec.WriteEndElement();
      dec.WriteEndElement();			
      dec.Flush();
      if (host != "") {
	  		WebClient wc = new  WebClient();
	  		try {
	      	CompilerOutput.Message("Sending compiled to " + host);
	      	wc.Headers.Add("User-Agent: BoPi Client");
	      	wc.Headers.Add("Content-Type: text/xml; charset=utf-8");
	      	wc.Headers.Add("SOAPAction: http://cs.unibo.it/BoPi/loadcode");
	      	wc.UploadData(host, "POST", System.Text.Encoding.UTF8.GetBytes(XmlUtil.AddSoap(swrdec.ToString())));
	      	CompilerOutput.Message("Success sending compiled to " + host);
	    	} catch (Exception) {
	      	CompilerOutput.Message("Failed sending compiled to "+host);
	    	}
			}
    }
  }
}
